package org.globus.exec.utils.audit;

import java.util.StringTokenizer;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.log4j.jdbc.JDBCAppender;
import org.apache.log4j.spi.LoggingEvent;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.ResultSet;

import org.globus.exec.utils.Resources;
import org.globus.util.I18n;


public class AuditDatabaseAppender extends JDBCAppender
{
    static Log logger = LogFactory.getLog(AuditDatabaseAppender.class);
    private static I18n i18n = I18n.getI18n(Resources.class.getName());

    private static String DB_DRIVER_CLASS = null;
    private static String DB_URL = null;
    private static String DB_USER = null;
    private static String DB_PASSWORD = null;
    private static String[] TABLE_COLUMN_NAMES = new String[] {
        "job_grid_id",
        "local_job_id",
        "subject_name",
        "username",
        "idempotence_id",
        "creation_time",
        "queued_time",
        "stage_in_grid_id",
        "stage_out_grid_id",
        "clean_up_grid_id",
        "globus_toolkit_version",
        "resource_manager_type",
        "job_description",
        "success_flag",
        "finished_flag"
    };
    private static String BASE_SQL;

    private boolean instanceInitialized = false;

    public AuditDatabaseAppender() { }

    public static void initialize(
        String                              dbDriverClass,
        String                              dbUrl,
        String                              dbUser,
        String                              dbPassowrd)
    {
        DB_DRIVER_CLASS = dbDriverClass;
        DB_URL = dbUrl;
        DB_USER = dbUser;
        DB_PASSWORD = dbPassowrd;

        StringBuffer sql = new StringBuffer("INSERT INTO gram_audit_table (");
        for (int index=0; index<TABLE_COLUMN_NAMES.length; index++)
        {
            sql.append(TABLE_COLUMN_NAMES[index]);
            if (index < (TABLE_COLUMN_NAMES.length-1))
            {
                sql.append(",");
            }
        }
        sql.append(") values (");

        BASE_SQL = sql.toString();
    }

    public static String format(
        String                              unformatedDatum)
    {
        return "\"" + escape(unformatedDatum) + "\"";
    }

    private static String escape(
        String                              unescapedString)
    {
        return unescapedString.replaceAll("\"", "&quot;");
    }

    private static String unescape(
        String                              escapedString)
    {
        return escapedString.replaceAll("&quot;", "\"");
    }

    public void append(
        LoggingEvent                        event)
    {

        if (!instanceInitialized)
        {
            if (   (DB_DRIVER_CLASS == null)
                && (DB_URL == null)
                && (DB_USER == null)
                && (DB_PASSWORD == null))
            {
                throw new RuntimeException(i18n.getMessage(
                    Resources.DATABASE_APPENDER_NOT_INITIALIZED));
            }

            setDriver(DB_DRIVER_CLASS);
            setURL(DB_URL);
            setUser(DB_USER);
            setPassword(DB_PASSWORD);

            instanceInitialized = true;
        }

        String message = event.getMessage().toString();
        logger.debug("Message: " + message);

        /*Parse message for column data.

          Format: "QUOTE_ESCAPED_STRING"[,"QUOTED_ESCAPED_STRING"]*

          QUOTE_ESCAPED_STRING is a string that has all it's double quote
          characters substituted for '&quot;'.
        */
        StringBuffer sql = new StringBuffer(BASE_SQL);
        String dbEntries = message.substring(8); // will get everything after 'insert/update: '
        logger.debug("DB Entries: " + dbEntries);

        StringTokenizer stok = new StringTokenizer(dbEntries, "\",\"");
        String token = null;
        if (message.indexOf("insert") == 0) {
            sql = new StringBuffer(BASE_SQL);
            while (stok.hasMoreTokens())
            {
                token = stok.nextToken();
                logger.debug("next token: " + token);
                                                                                                            
                                                                                                            
                if (token.equals("NULL"))
                {
                    sql.append("null");
                }
                else
                {
                    sql.append("'").append(unescape(token)).append("'");
                }
                                                                                                            
                                                                                                            
                if (stok.hasMoreTokens())
                {
                    sql.append(",");
                }
            }
            sql.append(");");
        } else { // update
            sql = new StringBuffer("UPDATE gram_audit_table SET ");
            String jobGridId = stok.nextToken();
            int index=1; // need to start right after job_grid_id
            while (stok.hasMoreTokens())
            {
                token = stok.nextToken();
                logger.debug("next token: " + token);
                sql.append(TABLE_COLUMN_NAMES[index]).append("=");
                if (token.equals("NULL"))
                {
                    sql.append("null");
                }
                else
                {
                    sql.append("'").append(unescape(token)).append("'");
                }
                                                                                                            
                if (stok.hasMoreTokens())
                {
                    sql.append(",");
                }
                index++;
            }
            sql.append(" WHERE job_grid_id=").append("'").append(unescape(jobGridId)).append("';");
        }
        logger.debug("SQL: " + sql.toString());
        setSql(sql.toString());

        super.append(event);
    }

    public static void main(String[] args)
    {
        String driverClass = "org.postgresql.Driver";
        String url = null;
        String username = null;
        String password = null;
        boolean checkSuccess = false;
        int index = 0;
        String help =
"globus-gram-append-audit-records [OPTIONS]\n" +
"Options:\n" +
" -d DRIVER-CLASS             Use the java class DRIVER-CLASS as the database\n" +
"                             driver\n" +
" -h | --help                 Print this help message\n"+
" -c CONFIG-FILENAME          Use CONFIG-FILENAME as a configuration file\n"+
" -n USERNAME                 Use USERNAME as the database owner\n"+
" -u URL                      Use URL as the database URL\n"+
" -p PASSWORD                 Use PASSWORD as the database password for USERNAME\n"+
" --check                     Perform checks that the database action succeed\n"+
"\n"+
"AUDIT-RECORD: A comma-delimited set of quoted strings, in the following order:\n"+
" JOB_GRID_ID                 An encoded EPR or Job Contact string for the job resource\n"+
" LOCAL_JOB_ID                The local scheduler's ID for the job\n"+
" SUBJECT_NAME                Subject name of the credential used to submit the job\n"+
" USERNAME                    Local username that the job is running as.\n"+
" IDEMPOTENCE_ID              XXX\n"+
" CREATION_TIME               Time when the Job Manager resource was created (in format\n"+
"                             YYYY MON HH:MM:SS TZ YEAR)\n"+
" QUEUED_TIME                 Time when the job was queued into the local\n"+
"                             scheduler (in format YYYY MON HH:MM:SS TZ YEAR)\n"+
" STAGE_IN_GRID_ID            An encoded EPR of the RFT resource used for file stage in\n"+
" STAGE_OUT_GRID_ID           An encoded EPR of the RFT resource used for file\n"+
"                             stage out.\n"+
" CLEAN_UP_GRID_ID            An encoded EPR of the RFT resource used for file\n"+
"                             clean up.\n"+
" GLOBUS_TOOLKIT_VERSION      A string containing the toolkit version\n"+
" RESOURCE_MANAGER_TYPE       The scheduler type used to execute the job\n"+
" JOB_DESCRIPTION             The RSL or JobDescription of the user's job\n"+
" SUCCESS_FLAG                A boolean value indicating\n"+
"                             the job completed successfully (\"true\"), or failed\n"+
"                             (\"false\").\n"+
" FINISHED_FLAG               A boolean value indicating\n"+
"                             the job has finished executing.\n"+
"CONFIG-FILENAME: A filename pointing to a property file which contains configuration\n"+
"properties. Valid properties are: URL, DRIVERCLASS, USERNAME, and PASSWORD.\n";

        try
        {
            for (index = 0; index < args.length; index++)
            {
                if (args[index].equals("-d"))
                {
                    //driver class
                    driverClass = args[++index];
                }
                else if (args[index].equals("-u"))
                {
                    //URL
                    url = args[++index];
                }
                else if (args[index].equals("-n"))
                {
                    //username
                    username = args[++index];
                }
                else if (args[index].equals("-p"))
                {
                    //password
                    password = args[++index];
                }
                else if (args[index].equals("--check"))
                {
                    checkSuccess = true;
                }
                else if (args[index].equals("-c"))
                {
                    java.io.File configFile = new java.io.File(args[++index]);
                    java.util.Properties prop = new java.util.Properties();
                    prop.load(new java.io.FileInputStream(configFile));

                    String p =  prop.getProperty("USERNAME");

                    if (p != null)
                    {
                        username = p;
                    }

                    p = prop.getProperty("DRIVERCLASS");

                    if (p != null)
                    {
                        driverClass = p;
                    }

                    p = prop.getProperty("URL");
                    if (p != null)
                    {
                        url = p;
                    }

                    p = prop.getProperty("PASSWORD");

                    if (p != null)
                    {
                        password = p;
                    }
                }
                else if (args[index].equals("-h")
                        || args[index].equals("-help")
                        || args[index].equals("--help"))
                {
                    System.out.println(help);
                    System.exit(0);
                }
                else
                {
                    System.err.println("Unknown option " + args[index]);
                    System.exit(1);
                }
            }
        }
        catch (ArrayIndexOutOfBoundsException e)
        {
            System.err.println("Missing option to " + args[index-1]);
            System.exit(1);
        }
        catch (java.io.IOException e)
        {
            System.err.println("Unable to read config file.");
            System.exit(1);
        }

        if (driverClass == null)
        {
            System.err.println("Error: driverClass not specified");
            System.exit(1);
        }
        else if (url == null)
        {
            System.err.println("Error: url not specified");
            System.exit(1);
        }
        else if (username == null)
        {
            System.err.println("Error: username not specified");
            System.exit(1);
        }
        else if (password == null)
        {
            System.err.println("Error: password not specified");
            System.exit(1);
        }
        logger.debug(driverClass);
        logger.debug(url);
        logger.debug(username);
        logger.debug(password);

        
        
        // check if database connection can be established
        // with the given values (if desired)
        Connection connection =  null;
        Statement statement = null;
        if (checkSuccess)
        {         
          try
          {
             Class.forName(driverClass);
             connection = DriverManager.getConnection(
                              url, username, password);
          }   
          catch (ClassNotFoundException e)
          {
             System.err.println("Can't load database driver "+driverClass);
             System.exit(1); 
          }
          catch (SQLException e) {
             System.err.println("Database Connection Error");
             System.exit(1);       
          }
        }
        
        initialize(driverClass, url, username, password);

        Log auditLogger = LogFactory.getLog(
            AuditDatabaseAppender.class.getName() + ".audit");
        java.io.BufferedReader r = new java.io.BufferedReader(
                new java.io.InputStreamReader(System.in));
        String record;

        try
        {
            while ((record = r.readLine()) != null)
            {
                record = record.replaceAll("'", "''");
                
                auditLogger.info("insert: " + record);
                
                // check if a record for that grid-id was inserted (if desired)
                if (checkSuccess)
                {
                   String jobGridId = "";
                   StringTokenizer st = new StringTokenizer(record,"\",\"");
                  
                   if (st.hasMoreTokens())
                   {
                      jobGridId = st.nextToken();
                      jobGridId.replaceAll("\"","");
                   }

                   statement = connection.createStatement();
                   ResultSet result = statement.executeQuery (
                     "SELECT count(*) from gram_audit_table WHERE job_grid_id='"+
                     jobGridId+"'");

                   if (result.next())
                   {
                       int count = result.getInt(1);
                       if (count < 1)
                       {
                           System.err.println(
                              "No audit record for that job in the database");
                           System.exit(1);
                       }
                       else if (count > 1)
                       {
                           System.err.println(
                              "More than one audit record for that job in the database");
                           System.exit(1);
                       }
                   } 
                   else
                   {
                      System.err.println(
                         "No query of the database for that job possible");
                      System.exit(1);
                   }

                }
            }
            
            if (checkSuccess)
            {
              try
              {
                 connection.close();
              }
              catch (Exception e)
              {}
            }  
            
            System.exit(0);
        }
        catch(java.io.IOException e)
        {
            System.err.println("Error reading audit record");
            System.exit(1);
        }
        catch (SQLException e)
        {
             System.err.println(
                 "Error while checking audit record in database");
                  e.printStackTrace();
             System.exit(1);
        }
        finally
        {
           if (checkSuccess)
           {
             try
             {
                connection.close();
             }
             catch (Exception e) 
             {}
           }   
        }
    }
}
