package org.globus.exec.utils.audit;

import java.security.MessageDigest;
import java.util.StringTokenizer;

import javax.xml.namespace.QName;

import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xml.security.utils.Base64;

import org.globus.exec.utils.ManagedJobConstants;
import org.globus.exec.utils.Resources;
import org.globus.transfer.reliable.service.RFTConstants;
import org.globus.util.I18n;
import org.globus.wsrf.encoding.ObjectDeserializer;

import org.w3c.dom.Element;

public class AuditUtil
{
    static Log logger = LogFactory.getLog(AuditUtil.class);
    private static I18n i18n = I18n.getI18n(Resources.class.getName());

    private static final String DELIMITER = "?";
    private static final String DIGEST_ALGORITHM = "SHA";
    private static final Class RESOURCE_KEY_CLASS = String.class;

    /** Returns a string representation of EPR using the resource key.
     *  The format used is To Address + "?" + hash(resourceKey).
     *  Code borrowed from org.globus.wsrf.impl.security.util.EPRUtil in HEAD.
     *  This method should become a wrapper around EPRUtil once it is included
     *  in the same branch.
     */
    public static String eprToGridId(
        EndpointReferenceType               epr)
        throws                              Exception
    {
        MessageElement resourceKey = epr.getProperties().get(
            ManagedJobConstants.RESOURCE_KEY_QNAME);

        if (resourceKey == null)
        {
            resourceKey = epr.getProperties().get(
                RFTConstants.RFT_KEY_QNAME);
        }

        if (resourceKey == null)
        {
            throw new Exception(i18n.getMessage(
                Resources.SUPPORTED_RESOURCE_KEY_NOT_FOUND,
                ManagedJobConstants.RESOURCE_KEY_QNAME.toString()
                + ", " + RFTConstants.RFT_KEY_QNAME.toString()));
        }

        return eprToGridId(epr, resourceKey);
    }

    public static String eprToGridId(
        EndpointReferenceType               epr,
        MessageElement                      resourceKey)
        throws                              Exception
    {
        if (epr == null)
        {
            return null;
        }

        String toAddress = epr.getAddress().toString();

        Element resourceKeyElement = resourceKey.getAsDOM();

        String key = (String)ObjectDeserializer.toObject(
            resourceKeyElement,
            RESOURCE_KEY_CLASS);

        MessageDigest mda = MessageDigest.getInstance(DIGEST_ALGORITHM);
        mda.update(key.getBytes());
        byte[] digest = mda.digest();
        String keyDigest = Base64.encode(digest);

        return toAddress + DELIMITER + keyDigest;
    }
}
